@@ -12,12 +12,12 @@
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
-  | Authors: Antony Dovgal <tony2001@phpclub.net>                        |
+  | Authors: Antony Dovgal <tony@daylessday.org>                         |
   |          Mikael Johansson <mikael AT synd DOT info>                  |
   +----------------------------------------------------------------------+
 */
 
-/* $Id: memcache.c,v 1.98 2008/02/05 19:46:16 mikl Exp $ */
+/* $Id: memcache.c 327750 2012-09-23 01:47:59Z hradtke $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -42,6 +42,10 @@
 #include "php_memcache.h"
 #include "memcache_queue.h"
 
+#if HAVE_MEMCACHE_SESSION
+#include "ext/session/php_session.h"
+#endif
+
 #ifndef ZEND_ENGINE_2
 #define OnUpdateLong OnUpdateInt
 #endif
@@ -74,6 +78,7 @@ zend_function_entry memcache_functions[] = {
 	PHP_FE(memcache_decrement,		NULL)
 	PHP_FE(memcache_close,			NULL)
 	PHP_FE(memcache_flush,			NULL)
+	PHP_FE(memcache_setoptimeout,	NULL)
 	{NULL, NULL, NULL}
 };
 
@@ -96,6 +101,7 @@ static zend_function_entry php_memcache_class_functions[] = {
 	PHP_FALIAS(decrement,		memcache_decrement,			NULL)
 	PHP_FALIAS(close,			memcache_close,				NULL)
 	PHP_FALIAS(flush,			memcache_flush,				NULL)
+	PHP_FALIAS(setoptimeout,	memcache_setoptimeout,		NULL)
 	{NULL, NULL, NULL}
 };
 
@@ -187,6 +193,20 @@ static PHP_INI_MH(OnUpdateHashFunction) /* {{{ */
 }
 /* }}} */
 
+static PHP_INI_MH(OnUpdateDefaultTimeout) /* {{{ */
+{
+	long int lval;
+
+	lval = strtol(new_value, NULL, 10);
+	if (lval <= 0) {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "memcache.default_timeout must be a positive number greater than or equal to 1 ('%s' given)", new_value);
+		return FAILURE;
+	}
+	MEMCACHE_G(default_timeout_ms) = lval;
+	return SUCCESS;
+}
+/* }}} */
+
 /* {{{ PHP_INI */
 PHP_INI_BEGIN()
 	STD_PHP_INI_ENTRY("memcache.allow_failover",	"1",		PHP_INI_ALL, OnUpdateLong,		allow_failover,	zend_memcache_globals,	memcache_globals)
@@ -195,6 +215,7 @@ PHP_INI_BEGIN()
 	STD_PHP_INI_ENTRY("memcache.chunk_size",		"8192",		PHP_INI_ALL, OnUpdateChunkSize,	chunk_size,		zend_memcache_globals,	memcache_globals)
 	STD_PHP_INI_ENTRY("memcache.hash_strategy",		"standard",	PHP_INI_ALL, OnUpdateHashStrategy,	hash_strategy,	zend_memcache_globals,	memcache_globals)
 	STD_PHP_INI_ENTRY("memcache.hash_function",		"crc32",	PHP_INI_ALL, OnUpdateHashFunction,	hash_function,	zend_memcache_globals,	memcache_globals)
+	STD_PHP_INI_ENTRY("memcache.default_timeout_ms",	"1000",	PHP_INI_ALL, OnUpdateDefaultTimeout,	default_timeout_ms,	zend_memcache_globals,	memcache_globals)
 PHP_INI_END()
 /* }}} */
 
@@ -238,6 +259,7 @@ static void php_memcache_init_globals(zend_memcache_globals *memcache_globals_p
 	MEMCACHE_G(compression_level) = Z_DEFAULT_COMPRESSION;
 	MEMCACHE_G(hash_strategy)	  = MMC_STANDARD_HASH;
 	MEMCACHE_G(hash_function)	  = MMC_HASH_CRC32;
+	MEMCACHE_G(default_timeout_ms)= (MMC_DEFAULT_TIMEOUT) * 1000;
 }
 /* }}} */
 
@@ -302,7 +324,7 @@ PHP_MINFO_FUNCTION(memcache)
 	php_info_print_table_header(2, "memcache support", "enabled");
 	php_info_print_table_row(2, "Active persistent connections", buf);
 	php_info_print_table_row(2, "Version", PHP_MEMCACHE_VERSION);
-	php_info_print_table_row(2, "Revision", "$Revision: 1.98 $");
+	php_info_print_table_row(2, "Revision", "$Revision: 327750 $");
 	php_info_print_table_end();
 
 	DISPLAY_INI_ENTRIES();
@@ -332,6 +354,18 @@ void mmc_debug(const char *format, ...) /* {{{ */
 /* }}} */
 #endif
 
+static struct timeval _convert_timeoutms_to_ts(long msecs) /* {{{ */
+{
+	struct timeval tv;
+	int secs = 0;
+
+	secs = msecs / 1000;
+	tv.tv_sec = secs;
+	tv.tv_usec = ((msecs - (secs * 1000)) * 1000) % 1000000;
+	return tv;
+}
+/* }}} */
+
 static void _mmc_pool_list_dtor(zend_rsrc_list_entry *rsrc TSRMLS_DC) /* {{{ */
 {
 	mmc_pool_free((mmc_pool_t *)rsrc->ptr TSRMLS_CC);
@@ -405,7 +439,7 @@ static void mmc_server_sleep(mmc_t *mmc TSRMLS_DC) /*
 	mmc->failure_callback = NULL;
 	
 	if (mmc->error != NULL) {
-		efree(mmc->error);
+		pefree(mmc->error, mmc->persistent);
 		mmc->error = NULL;
 	}
 }
@@ -440,10 +474,10 @@ static void mmc_server_seterror(mmc_t *mmc, const char *error, int errnum) /* {{
 {
 	if (error != NULL) {
 		if (mmc->error != NULL) {
-			efree(mmc->error);
+			pefree(mmc->error, mmc->persistent);
 		}
 		
-		mmc->error = estrdup(error);
+		mmc->error = pestrdup(error, mmc->persistent);
 		mmc->errnum = errnum;
 	}
 }
@@ -464,8 +498,7 @@ static void mmc_server_received_error(mmc_t *mmc, int response_len)  /* {{{ */
 }
 /* }}} */
 
-int mmc_server_failure(mmc_t *mmc TSRMLS_DC) /*
-	determines if a request should be retried or is a hard network failure {{{ */
+int mmc_server_failure(mmc_t *mmc TSRMLS_DC) /*determines if a request should be retried or is a hard network failure {{{ */
 {
 	switch (mmc->status) {
 		case MMC_STATUS_DISCONNECTED:
@@ -482,14 +515,20 @@ int mmc_server_failure(mmc_t *mmc TSRMLS_DC) /*
 }
 /* }}} */
 
-static int mmc_server_store(mmc_t *mmc, const char *request, int request_len TSRMLS_DC) { /* {{{ */
+static int mmc_server_store(mmc_t *mmc, const char *request, int request_len TSRMLS_DC) /* {{{ */
+{
 	int response_len;
+	php_netstream_data_t *sock = (php_netstream_data_t*)mmc->stream->abstract;
+
+	if (mmc->timeoutms > 1) {
+		sock->timeout = _convert_timeoutms_to_ts(mmc->timeoutms);
+	}
 
 	if (php_stream_write(mmc->stream, request, request_len) != request_len) {
 		mmc_server_seterror(mmc, "Failed sending command and value to stream", 0);
 		return -1;
 	}
-	
+
 	if ((response_len = mmc_readline(mmc TSRMLS_CC)) < 0) {
 		return -1;
 	}
@@ -544,7 +583,6 @@ int mmc_prepare_key(zval *key, char *result, unsigned int *result_len TSRMLS_DC)
 
 		*keytmp = *key;
 		zval_copy_ctor(keytmp);
-		INIT_PZVAL(keytmp);
 		convert_to_string(keytmp);
 
 		res = mmc_prepare_key_ex(Z_STRVAL_P(keytmp), Z_STRLEN_P(keytmp), result, result_len TSRMLS_CC);
@@ -557,8 +595,7 @@ int mmc_prepare_key(zval *key, char *result, unsigned int *result_len TSRMLS_DC)
 }
 /* }}} */
 
-static unsigned int mmc_hash_crc32(const char *key, int key_len) /* 
-	CRC32 hash {{{ */
+static unsigned int mmc_hash_crc32(const char *key, int key_len) /* CRC32 hash {{{ */
 {
 	unsigned int crc = ~0;
 	int i;
@@ -571,8 +608,7 @@ static unsigned int mmc_hash_crc32(const char *key, int key_len) /*
 }
 /* }}} */
 
-static unsigned int mmc_hash_fnv1a(const char *key, int key_len) /* 
-	FNV-1a hash {{{ */
+static unsigned int mmc_hash_fnv1a(const char *key, int key_len) /* FNV-1a hash {{{ */
 {
 	unsigned int hval = FNV_32_INIT;
 	int i;
@@ -580,9 +616,9 @@ static unsigned int mmc_hash_fnv1a(const char *key, int key_len) /*
 	for (i=0; i<key_len; i++) {
 		hval ^= (unsigned int)key[i];
 		hval *= FNV_32_PRIME;
-    }
+	}
 
-    return hval;
+	return hval;
 }
 /* }}} */
 
@@ -675,8 +711,7 @@ void mmc_pool_add(mmc_pool_t *pool, mmc_t *mmc, unsigned int weight) /* {{{ */
 }
 /* }}} */
 
-static int mmc_pool_close(mmc_pool_t *pool TSRMLS_DC) /* 
-	disconnects and removes all servers in the pool {{{ */
+static int mmc_pool_close(mmc_pool_t *pool TSRMLS_DC) /* disconnects and removes all servers in the pool {{{ */
 {
 	if (pool->num_servers) {
 		int i;
@@ -726,7 +761,7 @@ int mmc_pool_store(mmc_pool_t *pool, const char *command, int command_len, const
 		unsigned long data_len;
 
 		if (!mmc_compress(&data, &data_len, value, value_len TSRMLS_CC)) {
-			mmc_server_seterror(mmc, "Failed to compress data", 0);
+			/* mmc_server_seterror(mmc, "Failed to compress data", 0); */
 			return -1;
 		}
 
@@ -742,32 +777,32 @@ int mmc_pool_store(mmc_pool_t *pool, const char *command, int command_len, const
 		}
 	}
 
-	request = emalloc(
-		command_len
-		+ 1				/* space */
-		+ key_len
-		+ 1				/* space */
-		+ MAX_LENGTH_OF_LONG
-		+ 1 			/* space */
-		+ MAX_LENGTH_OF_LONG
-		+ 1 			/* space */
-		+ MAX_LENGTH_OF_LONG
-		+ sizeof("\r\n") - 1
-		+ value_len
-		+ sizeof("\r\n") - 1
-		+ 1
-		);
+	request = emalloc( 		
+			command_len 		
+			+ 1 /* space */ 		
+			+ key_len 		
+			+ 1 /* space */ 		
+			+ MAX_LENGTH_OF_LONG 		
+			+ 1 /* space */ 		
+			+ MAX_LENGTH_OF_LONG 		
+			+ 1 /* space */ 		
+			+ MAX_LENGTH_OF_LONG 		
+			+ sizeof("\r\n") - 1 		
+			+ value_len 		
+			+ sizeof("\r\n") - 1 		
+			+ 1 		
+			); 		
 
 	request_len = sprintf(request, "%s %s %d %d %d\r\n", command, key, flags, expire, value_len);
 
-	memcpy(request + request_len, value, value_len);
-	request_len += value_len;
+	memcpy(request + request_len, value, value_len); 		
+	request_len += value_len; 		
 
-	memcpy(request + request_len, "\r\n", sizeof("\r\n") - 1);
-	request_len += sizeof("\r\n") - 1;
+	memcpy(request + request_len, "\r\n", sizeof("\r\n") - 1); 		
+	request_len += sizeof("\r\n") - 1; 		
 
 	request[request_len] = '\0';
-
+	
 	while (result < 0 && (mmc = mmc_pool_find(pool, key, key_len TSRMLS_CC)) != NULL) {
 		if ((result = mmc_server_store(mmc, request, request_len TSRMLS_CC)) < 0) {
 			mmc_server_failure(mmc TSRMLS_CC);
@@ -861,14 +896,14 @@ static int mmc_get_pool(zval *id, mmc_pool_t **pool TSRMLS_DC) /* {{{ */
 	int resource_type;
 
 	if (Z_TYPE_P(id) != IS_OBJECT || zend_hash_find(Z_OBJPROP_P(id), "connection", sizeof("connection"), (void **) &connection) == FAILURE) {
-		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failed to extract 'connection' variable from object");
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "No servers added to memcache connection");
 		return 0;
 	}
 
 	*pool = (mmc_pool_t *) zend_list_find(Z_LVAL_PP(connection), &resource_type);
 
 	if (!*pool || resource_type != le_memcache_pool) {
-		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unknown connection identifier");
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid Memcache->connection member variable");
 		return 0;
 	}
 
@@ -887,8 +922,12 @@ static int _mmc_open(mmc_t *mmc, char **error_string, int *errnum TSRMLS_DC) /*
 		mmc_server_disconnect(mmc TSRMLS_CC);
 	}
 
-	tv.tv_sec = mmc->timeout;
-	tv.tv_usec = 0;
+	if (mmc->connect_timeoutms > 0) {
+		tv = _convert_timeoutms_to_ts(mmc->connect_timeoutms);
+	} else {
+		tv.tv_sec = mmc->timeout;
+		tv.tv_usec = 0;
+	}
 
 	if (mmc->port) {
 		hostname_len = spprintf(&hostname, 0, "%s:%d", mmc->host, mmc->port);
@@ -960,7 +999,7 @@ static int _mmc_open(mmc_t *mmc, char **error_string, int *errnum TSRMLS_DC) /*
 	mmc->status = MMC_STATUS_CONNECTED;
 
 	if (mmc->error != NULL) {
-		efree(mmc->error);
+		pefree(mmc->error, mmc->persistent);
 		mmc->error = NULL;
 	}
 
@@ -1018,8 +1057,7 @@ static void mmc_server_disconnect(mmc_t *mmc TSRMLS_DC) /* {{{ */
 }
 /* }}} */
 
-void mmc_server_deactivate(mmc_t *mmc TSRMLS_DC) /* 
-	disconnect and marks the server as down {{{ */
+void mmc_server_deactivate(mmc_t *mmc TSRMLS_DC) /* 	disconnect and marks the server as down {{{ */
 {
 	mmc_server_disconnect(mmc TSRMLS_CC);
 	mmc->status = MMC_STATUS_FAILED;
@@ -1028,7 +1066,13 @@ void mmc_server_deactivate(mmc_t *mmc TSRMLS_DC) /*
 	if (mmc->failure_callback != NULL) {
 		zval *retval = NULL;
 		zval *host, *tcp_port, *udp_port, *error, *errnum;
-		zval **params[5] = {&host, &tcp_port, &udp_port, &error, &errnum};
+		zval **params[5];
+
+		params[0] = &host;
+		params[1] = &tcp_port;
+		params[2] = &udp_port;
+		params[3] = &error;
+		params[4] = &errnum;
 
 		MAKE_STD_ZVAL(host);
 		MAKE_STD_ZVAL(tcp_port); MAKE_STD_ZVAL(udp_port);
@@ -1125,6 +1169,7 @@ static int mmc_sendcmd(mmc_t *mmc, const char *cmd, int cmdlen TSRMLS_DC) /* {{{
 {
 	char *command;
 	int command_len;
+	php_netstream_data_t *sock = (php_netstream_data_t*)mmc->stream->abstract;
 
 	if (!mmc || !cmd) {
 		return -1;
@@ -1138,6 +1183,10 @@ static int mmc_sendcmd(mmc_t *mmc, const char *cmd, int cmdlen TSRMLS_DC) /* {{{
 	command_len = cmdlen + sizeof("\r\n") - 1;
 	command[command_len] = '\0';
 
+	if (mmc->timeoutms > 1) {
+		sock->timeout = _convert_timeoutms_to_ts(mmc->timeoutms);
+	}
+
 	if (php_stream_write(mmc->stream, command, command_len) != command_len) {
 		mmc_server_seterror(mmc, "Failed writing command to stream", 0);
 		efree(command);
@@ -1212,7 +1261,8 @@ static int mmc_postprocess_value(zval **return_value, char *value, int value_len
 		ZVAL_FALSE(*return_value);
 		PHP_VAR_UNSERIALIZE_DESTROY(var_hash);
 		efree(value);
-		return -1;
+		php_error_docref(NULL TSRMLS_CC, E_NOTICE, "unable to unserialize data");
+		return 0;
 	}
 
 	PHP_VAR_UNSERIALIZE_DESTROY(var_hash);
@@ -1426,7 +1476,8 @@ static int mmc_read_value(mmc_t *mmc, char **key, int *key_len, char **value, in
 				efree(*key);
 			}
 			efree(data);
-			return -1;
+			php_error_docref(NULL TSRMLS_CC, E_NOTICE, "unable to uncompress data");
+			return 0;
 		}
 
 		efree(data);
@@ -1732,24 +1783,24 @@ static int mmc_incr_decr(mmc_t *mmc, int cmd, char *key, int key_len, int value,
 static void php_mmc_store(INTERNAL_FUNCTION_PARAMETERS, char *command, int command_len) /* {{{ */
 {
 	mmc_pool_t *pool;
-	zval *var, *mmc_object = getThis();
+	zval *value, *mmc_object = getThis();
 
-	int result, value_len, key_len;
-	char *value, *key;
+	int result, key_len;
+	char *key;
 	long flags = 0, expire = 0;
 	char key_tmp[MMC_KEY_MAX_SIZE];
 	unsigned int key_tmp_len;
 
-	php_serialize_data_t var_hash;
+	php_serialize_data_t value_hash;
 	smart_str buf = {0};
 
 	if (mmc_object == NULL) {
-		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "Osz|ll", &mmc_object, memcache_class_entry_ptr, &key, &key_len, &var, &flags, &expire) == FAILURE) {
+		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "Osz|ll", &mmc_object, memcache_class_entry_ptr, &key, &key_len, &value, &flags, &expire) == FAILURE) {
 			return;
 		}
 	}
 	else {
-		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sz|ll", &key, &key_len, &var, &flags, &expire) == FAILURE) {
+		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "sz|ll", &key, &key_len, &value, &flags, &expire) == FAILURE) {
 			return;
 		}
 	}
@@ -1762,48 +1813,58 @@ static void php_mmc_store(INTERNAL_FUNCTION_PARAMETERS, char *command, int comma
 		RETURN_FALSE;
 	}
 
-	switch (Z_TYPE_P(var)) {
+	switch (Z_TYPE_P(value)) {
 		case IS_STRING:
-			value = Z_STRVAL_P(var);
-			value_len = Z_STRLEN_P(var);
+			result = mmc_pool_store(
+				pool, command, command_len, key_tmp, key_tmp_len, flags, expire, 
+				Z_STRVAL_P(value), Z_STRLEN_P(value) TSRMLS_CC);
 			break;
 
 		case IS_LONG:
 		case IS_DOUBLE:
-		case IS_BOOL:
-			convert_to_string(var);
-			value = Z_STRVAL_P(var);
-			value_len = Z_STRLEN_P(var);
-			break;
-
-		default: {
-			zval var_copy, *var_copy_ptr;
+		case IS_BOOL: {
+			zval value_copy;
 
 			/* FIXME: we should be using 'Z' instead of this, but unfortunately it's PHP5-only */
-			var_copy = *var;
-			zval_copy_ctor(&var_copy);
+			value_copy = *value;
+			zval_copy_ctor(&value_copy);
+			convert_to_string(&value_copy);
 
-			var_copy_ptr = &var_copy;
+			result = mmc_pool_store(
+				pool, command, command_len, key_tmp, key_tmp_len, flags, expire, 
+				Z_STRVAL(value_copy), Z_STRLEN(value_copy) TSRMLS_CC);
 
-			PHP_VAR_SERIALIZE_INIT(var_hash);
-			php_var_serialize(&buf, &var_copy_ptr, &var_hash TSRMLS_CC);
-			PHP_VAR_SERIALIZE_DESTROY(var_hash);
+			zval_dtor(&value_copy);
+			break;
+		}
+
+		default: {
+			zval value_copy, *value_copy_ptr;
+
+			/* FIXME: we should be using 'Z' instead of this, but unfortunately it's PHP5-only */
+			value_copy = *value;
+			zval_copy_ctor(&value_copy);
+			value_copy_ptr = &value_copy;
+
+			PHP_VAR_SERIALIZE_INIT(value_hash);
+			php_var_serialize(&buf, &value_copy_ptr, &value_hash TSRMLS_CC);
+			PHP_VAR_SERIALIZE_DESTROY(value_hash);
 
 			if (!buf.c) {
 				/* something went really wrong */
-				zval_dtor(&var_copy);
+				zval_dtor(&value_copy);
+				php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failed to serialize value");
 				RETURN_FALSE;
 			}
 
-			value = buf.c;
-			value_len = buf.len;
 			flags |= MMC_SERIALIZED;
-			zval_dtor(&var_copy);
-		}
-			break;
-	}
+			zval_dtor(&value_copy);
 
-	result = mmc_pool_store(pool, command, command_len, key_tmp, key_tmp_len, flags, expire, value, value_len TSRMLS_CC);
+			result = mmc_pool_store(
+				pool, command, command_len, key_tmp, key_tmp_len, flags, expire, 
+				buf.c, buf.len TSRMLS_CC);
+		}
+	}
 
 	if (flags & MMC_SERIALIZED) {
 		smart_str_free(&buf);
@@ -1867,12 +1928,16 @@ static void php_mmc_connect (INTERNAL_FUNCTION_PARAMETERS, int persistent) /* {{
 	mmc_pool_t *pool;
 	int resource_type, host_len, errnum = 0, list_id;
 	char *host, *error_string = NULL;
-	long port = MEMCACHE_G(default_port), timeout = MMC_DEFAULT_TIMEOUT;
+	long port = MEMCACHE_G(default_port), timeout = MMC_DEFAULT_TIMEOUT, timeoutms = 0;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|ll", &host, &host_len, &port, &timeout) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|lll", &host, &host_len, &port, &timeout, &timeoutms) == FAILURE) {
 		return;
 	}
 
+	if (timeoutms < 1) { 
+		timeoutms = MEMCACHE_G(default_timeout_ms);
+	}
+
 	/* initialize and connect server struct */
 	if (persistent) {
 		mmc = mmc_find_persistent(host, host_len, port, timeout, MMC_DEFAULT_RETRY TSRMLS_CC);
@@ -1882,6 +1947,9 @@ static void php_mmc_connect (INTERNAL_FUNCTION_PARAMETERS, int persistent) /* {{
 		mmc = mmc_server_new(host, host_len, port, 0, timeout, MMC_DEFAULT_RETRY TSRMLS_CC);
 	}
 
+	mmc->timeout = timeout;
+	mmc->connect_timeoutms = timeoutms;
+
 	if (!mmc_open(mmc, 1, &error_string, &errnum TSRMLS_CC)) {
 		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Can't connect to %s:%ld, %s (%d)", host, port, error_string ? error_string : "Unknown error", errnum);
 		if (!persistent) {
@@ -1899,7 +1967,7 @@ static void php_mmc_connect (INTERNAL_FUNCTION_PARAMETERS, int persistent) /* {{
 		mmc_pool_add(pool, mmc, 1);
 
 		object_init_ex(return_value, memcache_class_entry_ptr);
-		list_id = zend_list_insert(pool, le_memcache_pool);
+		list_id = MEMCACHE_LIST_INSERT(pool, le_memcache_pool);
 		add_property_resource(return_value, "connection", list_id);
 	}
 	else if (zend_hash_find(Z_OBJPROP_P(mmc_object), "connection", sizeof("connection"), (void **) &connection) != FAILURE) {
@@ -1916,7 +1984,7 @@ static void php_mmc_connect (INTERNAL_FUNCTION_PARAMETERS, int persistent) /* {{
 		pool = mmc_pool_new(TSRMLS_C);
 		mmc_pool_add(pool, mmc, 1);
 
-		list_id = zend_list_insert(pool, le_memcache_pool);
+		list_id = MEMCACHE_LIST_INSERT(pool, le_memcache_pool);
 		add_property_resource(mmc_object, "connection", list_id);
 		RETURN_TRUE;
 	}
@@ -1950,29 +2018,33 @@ PHP_FUNCTION(memcache_add_server)
 	zval **connection, *mmc_object = getThis(), *failure_callback = NULL;
 	mmc_pool_t *pool;
 	mmc_t *mmc;
-	long port = MEMCACHE_G(default_port), weight = 1, timeout = MMC_DEFAULT_TIMEOUT, retry_interval = MMC_DEFAULT_RETRY;
+	long port = MEMCACHE_G(default_port), weight = 1, timeout = MMC_DEFAULT_TIMEOUT, retry_interval = MMC_DEFAULT_RETRY, timeoutms = 0;
 	zend_bool persistent = 1, status = 1;
 	int resource_type, host_len, list_id;
 	char *host;
 
 	if (mmc_object) {
-		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|lblllbz", &host, &host_len, &port, &persistent, &weight, &timeout, &retry_interval, &status, &failure_callback) == FAILURE) {
+		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|lblllbzl", &host, &host_len, &port, &persistent, &weight, &timeout, &retry_interval, &status, &failure_callback, &timeoutms) == FAILURE) {
 			return;
 		}
 	}
 	else {
-		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "Os|lblllbz", &mmc_object, memcache_class_entry_ptr, &host, &host_len, &port, &persistent, &weight, &timeout, &retry_interval, &status, &failure_callback) == FAILURE) {
+		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "Os|lblllbzl", &mmc_object, memcache_class_entry_ptr, &host, &host_len, &port, &persistent, &weight, &timeout, &retry_interval, &status, &failure_callback, &timeoutms) == FAILURE) {
 			return;
 		}
 	}
 
-	if (weight < 0) {
+	if (timeoutms < 1) { 
+		timeoutms = MEMCACHE_G(default_timeout_ms);
+	}
+
+	if (weight < 1) {
 		php_error_docref(NULL TSRMLS_CC, E_WARNING, "weight must be a positive integer");
 		RETURN_FALSE;
 	}
 
 	if (failure_callback != NULL && Z_TYPE_P(failure_callback) != IS_NULL) {
-		if (!zend_is_callable(failure_callback, 0, NULL)) {
+		if (!MEMCACHE_IS_CALLABLE(failure_callback, 0, NULL)) {
 			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid failure callback");
 			RETURN_FALSE;
 		}
@@ -1987,6 +2059,8 @@ PHP_FUNCTION(memcache_add_server)
 		mmc = mmc_server_new(host, host_len, port, 0, timeout, retry_interval TSRMLS_CC);
 	}
 
+	mmc->connect_timeoutms = timeoutms;
+
 	/* add server in failed mode */
 	if (!status) {
 		mmc->status = MMC_STATUS_FAILED;
@@ -2000,7 +2074,7 @@ PHP_FUNCTION(memcache_add_server)
 	/* initialize pool if need be */
 	if (zend_hash_find(Z_OBJPROP_P(mmc_object), "connection", sizeof("connection"), (void **) &connection) == FAILURE) {
 		pool = mmc_pool_new(TSRMLS_C);
-		list_id = zend_list_insert(pool, le_memcache_pool);
+		list_id = MEMCACHE_LIST_INSERT(pool, le_memcache_pool);
 		add_property_resource(mmc_object, "connection", list_id);
 	}
 	else {
@@ -2056,7 +2130,7 @@ PHP_FUNCTION(memcache_set_server_params)
 	}
 
 	if (failure_callback != NULL && Z_TYPE_P(failure_callback) != IS_NULL) {
-		if (!zend_is_callable(failure_callback, 0, NULL)) {
+		if (!MEMCACHE_IS_CALLABLE(failure_callback, 0, NULL)) {
 			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid failure callback");
 			RETURN_FALSE;
 		}
@@ -2140,8 +2214,7 @@ mmc_t *mmc_find_persistent(char *host, int host_len, int port, int timeout, int
 	int hash_key_len;
 
 	MMC_DEBUG(("mmc_find_persistent: seeking for persistent connection"));
-	hash_key = emalloc(sizeof("mmc_connect___") - 1 + host_len + MAX_LENGTH_OF_LONG + 1);
-	hash_key_len = sprintf(hash_key, "mmc_connect___%s:%d", host, port);
+	hash_key_len = spprintf(&hash_key, 0, "mmc_connect___%s:%d", host, port);
 
 	if (zend_hash_find(&EG(persistent_list), hash_key, hash_key_len+1, (void **) &le) == FAILURE) {
 		zend_rsrc_list_entry new_le;
@@ -2156,7 +2229,7 @@ mmc_t *mmc_find_persistent(char *host, int host_len, int port, int timeout, int
 			mmc_server_free(mmc TSRMLS_CC);
 			mmc = NULL;
 		} else {
-			zend_list_insert(mmc, le_pmemcache);
+			MEMCACHE_LIST_INSERT(mmc, le_pmemcache);
 		}
 	}
 	else if (le->type != le_pmemcache || le->ptr == NULL) {
@@ -2174,7 +2247,7 @@ mmc_t *mmc_find_persistent(char *host, int host_len, int port, int timeout, int
 			mmc = NULL;
 		}
 		else {
-			zend_list_insert(mmc, le_pmemcache);
+			MEMCACHE_LIST_INSERT(mmc, le_pmemcache);
 		}
 	}
 	else {
@@ -2443,8 +2516,7 @@ PHP_FUNCTION(memcache_get_extended_stats)
 	for (i=0; i<pool->num_servers; i++) {
 		MAKE_STD_ZVAL(stats);
 
-		hostname = emalloc(strlen(pool->servers[i]->host) + MAX_LENGTH_OF_LONG + 1 + 1);
-		hostname_len = sprintf(hostname, "%s:%d", pool->servers[i]->host, pool->servers[i]->port);
+		hostname_len = spprintf(&hostname, 0, "%s:%d", pool->servers[i]->host, pool->servers[i]->port);
 
 		if (mmc_open(pool->servers[i], 1, NULL, NULL TSRMLS_CC)) {
 			if (mmc_get_stats(pool->servers[i], type, slabid, limit, stats TSRMLS_CC) < 0) {
@@ -2590,6 +2662,43 @@ PHP_FUNCTION(memcache_flush)
 }
 /* }}} */
 
+/* {{{ proto bool memcache_setoptimeout( object memcache , int timeoutms )
+   Set the timeout, in milliseconds, for subsequent operations on all open connections */
+PHP_FUNCTION(memcache_setoptimeout)
+{
+	mmc_pool_t *pool;
+	mmc_t *mmc;
+	int i;
+	zval *mmc_object = getThis();
+	long timeoutms = 0;
+
+	if (mmc_object == NULL) {
+		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "Ol", &mmc_object, memcache_class_entry_ptr, &timeoutms) == FAILURE) {
+			return;
+		}
+	}
+	else {
+		if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &timeoutms) == FAILURE) {
+			return;
+		}
+	}
+
+	if (timeoutms < 1) {
+		timeoutms = MEMCACHE_G(default_timeout_ms);
+	}
+
+	if (!mmc_get_pool(mmc_object, &pool TSRMLS_CC)) {
+		RETURN_FALSE;
+	}
+
+	for (i = 0; i < pool->num_servers; i++) {
+		mmc = pool->servers[i];
+		mmc->timeoutms = timeoutms;
+	}
+	RETURN_TRUE;
+}
+/* }}} */
+
 /*
  * Local variables:
  * tab-width: 4
